/**
 * Baobab Facet Class
 * ===================
 *
 * A class in charge of tree's computed data node.
 */
'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true
});

var _createClass = (function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ('value' in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; })();

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var _type = require('./type');

var _type2 = _interopRequireDefault(_type);

var _helpers = require('./helpers');

/**
 * Facet class
 *
 * @constructor
 * @param {Baobab}       tree       - The tree.
 * @param {array}        pathInTree - Path where the facets stands in its tree.
 * @param {array|object} definition - The facet's definition.
 */

var Facet = (function () {
  function Facet(tree, pathInTree, definition) {
    var _this = this;

    _classCallCheck(this, Facet);

    // Checking definition's type
    var definitionType = _type2['default'].facetDefinition(definition);

    // If the definition type is not valid, we cry
    if (!definitionType) throw (0, _helpers.makeError)('Baobab.Facet: attempting to create a computed data node with a ' + ('wrong definition (path: /' + pathInTree.join('/') + ').'), { path: pathInTree, definition: definition });

    // Properties
    this.tree = tree;
    this.computedData = null;
    this.type = definitionType;

    // Harmonizing
    if (definitionType === 'object') {
      this.getter = definition.get;
      this.projection = definition.cursors || {};
      this.paths = Object.keys(this.projection).map(function (k) {
        return _this.projection[k];
      });
    } else {
      this.getter = definition[definition.length - 1];
      this.projection = definition.slice(0, -1);
      this.paths = this.projection;
    }

    this._hasDynamicPaths = this.paths.some(function (p) {
      return _type2['default'].dynamicPath(p);
    });

    // Is the facet recursive?
    this.isRecursive = this.paths.some(function (p) {
      return !!_type2['default'].facetPath(p);
    });

    // Internal state
    this.state = {
      computed: false,
      killed: false
    };

    /**
     * Listener on the tree's `write` event.
     *
     * When the tree writes, this listener will check whether the updated paths
     * are of any use to the facet and, if so, will clean any computed data
     * so that the data may be recomputed when needed.
     */
    this.listener = function (_ref) {
      var path = _ref.data.path;

      if (_this.state.killed) return;

      // Is this facet affected by the current write event?
      var concerned = (0, _helpers.solveUpdate)([path], _this.relatedPaths());

      if (concerned) {
        _this.computedData = null;
        _this.state.computed = false;
      }
    };

    // Binding listener
    tree.on('write', this.listener);
  }

  /**
   * Method returning solved paths related to the facet.
   *
   * @return {array} - An array of related paths.
   */

  _createClass(Facet, [{
    key: 'relatedPaths',
    value: function relatedPaths() {
      var _this2 = this;

      var paths = undefined;

      if (this._hasDynamicPaths) paths = this.paths.map(function (p) {
        return (0, _helpers.getIn)(_this2.tree._data, p, _this2.tree._computedDataIndex).solvedPath;
      });else paths = this.paths;

      if (!this.isRecursive) return paths;else return paths.reduce(function (accumulatedPaths, path) {
        var facetPath = _type2['default'].facetPath(path);
        if (!facetPath) return accumulatedPaths.concat([path]);

        // Solving recursive path
        var relatedFacet = (0, _helpers.getIn)(_this2.tree._computedDataIndex, facetPath).data;

        return accumulatedPaths.concat(relatedFacet.relatedPaths());
      }, []);
    }

    /**
     * Getter method
     *
     * @return {mixed} - The facet's computed data.
     */
  }, {
    key: 'get',
    value: function get() {

      // Returning data if already computed
      if (this.state.computed) return this.computedData;

      // Computing data
      var deps = this.tree.project(this.projection),
          data = this.getter.apply(this.tree, this.type === 'object' ? [deps] : deps);

      this.state.computed = true;
      this.computedData = data;

      // If the tree is immutable, we need to freeze the data
      if (this.tree.options.immutable) (0, _helpers.deepFreeze)(this.computedData);

      return data;
    }

    /**
     * Method releasing the facet from memory.
     */
  }, {
    key: 'release',
    value: function release() {

      // Deleting properties
      delete this.getter;
      delete this.projection;
      delete this.paths;

      // Unbinding listener
      this.tree.off('write', this.listener);
      this.state.killed = true;
    }
  }]);

  return Facet;
})();

exports['default'] = Facet;
module.exports = exports['default'];